@@ -21,9 +21,10 @@ from sklearn.utils.multiclass import check_classification_targets
 from sklearn.utils.validation import check_array, check_is_fitted, check_random_state
 from statsmodels.distributions.empirical_distribution import ECDF
 
-from proset.objective import ClassifierObjective
+from proset.objectives.np_classifier_objective import NpClassifierObjective
+from proset.objectives.tf_classifier_objective import TfClassifierObjective
 from proset.set_manager import ClassifierSetManager
-from proset.shared import find_changes, check_feature_names, check_scale_offset, LOG_OFFSET
+import proset.shared as shared
 
 
 logger = logging.getLogger(__name__)
@@ -35,7 +36,6 @@ LOG_CAPTION = "  ".join(["{:>10s}"] * 6 + ["{:s}"]).format(
 LOG_MESSAGE = "  ".join(["{:10d}", "{:10d}", "{:10.1e}", "{:10.1e}", "{:10d}", "{:10d}", "{:s}"])
 
 LIMITED_M = 10  # parameters controlling L-BFGS-B fit
-LIMITED_FACTR = 1e7
 LIMITED_PGTOL = 1e-5
 LIMITED_MAXFUN = 15000
 LIMITED_MAXITER = 15000
@@ -56,6 +56,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
             alpha_w=0.95,
             num_candidates=1000,
             max_fraction=0.5,
+            solver_factr=1e7,
+            use_tensorflow=False,
             random_state=None
     ):
         """Initialize prototype set model with hyperparameters.
@@ -70,6 +72,11 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         :param num_candidates: positive integer; number of candidates for prototypes to try for each batch
         :param max_fraction: float in (0.0, 1.0); maximum fraction of candidates to draw from one group of candidates;
             candidates are grouped by class and whether the current model classifies them correctly or not
+        :param solver_factr: positive float; this is parameter factr of scipy.optimize.fmin_l_bfgs_b() which controls
+            when the algorithm stops because the change in objective function is too small; the documentation recommends
+            1e12 for low accuracy, 1e7 for moderate accuracy and 1e1 for extremely high accuracy
+        :param use_tensorflow: boolean; whether to use tensorflow for model fitting or rely on numpy only; scoring
+            always uses numpy only
         :param random_state: instance of np.random.RandomState, integer, or None; if a random state is passed, that
             state will be used for randomization; if an integer or None is passed, a new random state is generated using
             the argument as seed for every call to fit()
@@ -82,14 +89,16 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         self.num_candidates = num_candidates
         self.max_fraction = max_fraction
         self.random_state = random_state
+        self.use_tensorflow = use_tensorflow
+        self.solver_factr = solver_factr
 
     def fit(self, X, y, sample_weight=None, warm_start=False):
         """Fit proset model to data.
 
         :param X: 2D numpy float array; feature matrix; sparse matrices or infinite/missing values not supported
         :param y: list-like object; target for supervised learning
-        :param sample_weight: 1D numpy array of positive floats or None; sample weights used for likelihood calculation;
-            pass None to use unit weights
+        :param sample_weight: 1D numpy array of non-negative floats or None; sample weights used for likelihood
+            calculation; pass None to use unit weights
         :param warm_start: boolean; whether to create a new model or to add batches to an existing model
         :return: no return value; model updated in place
         """
@@ -98,7 +107,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         logger.info("Fit proset model with {} batches and penalties lambda_v = {:0.2e}, lambda_w = {:0.2e}".format(
             self.n_iter, self.lambda_v, self.lambda_w
         ))
-        MySetManager, MyObjective = self._get_compute_classes()  # pylint: disable=invalid-name
+        MySetManager, MyObjective = self._get_compute_classes(self.use_tensorflow)  # pylint: disable=invalid-name
         if not warm_start or not hasattr(self, "set_manager_"):
             self.set_manager_ = MySetManager(target=y)  # pylint: disable=attribute-defined-outside-init
         for i in range(self.n_iter):
@@ -121,7 +130,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
                 x0=starting_point,
                 bounds=bounds,
                 m=LIMITED_M,
-                factr=LIMITED_FACTR,
+                factr=self.solver_factr,
                 pgtol=LIMITED_PGTOL,
                 maxfun=LIMITED_MAXFUN,
                 maxiter=LIMITED_MAXITER,
@@ -153,6 +162,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
             raise TypeError("Parameter n_iter must be integer.")
         if self.n_iter < 0:
             raise ValueError("Parameter n_iter must not be negative.")
+        if self.solver_factr <= 0.0:
+            raise ValueError("Parameter solver_factr must be positive.")
         # validation of other parameters is left to the classes or functions relying on them
 
     # noinspection PyMethodMayBeStatic, PyUnresolvedReferences
@@ -172,26 +183,27 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         elif self.n_features_in_ != X.shape[1]:
             raise ValueError("Parameter X must have {} columns.".format(self.n_features_in_))
         if sample_weight is not None:
-            sample_weight = check_array(sample_weight, ensure_2d=False)
+            sample_weight = check_array(sample_weight, ensure_2d=False, **shared.FLOAT_TYPE)
             if sample_weight.shape[0] != X.shape[0]:
                 raise ValueError("Parameter sample_weight must have one element per row of X if not None.")
-        return X, self._validate_y(y, reset), sample_weight
+        return check_array(X, **shared.FLOAT_TYPE), self._validate_y(y, reset), sample_weight
 
     @abstractmethod
     def _validate_y(self, y, reset):  # pragma: no cover
         """Perform checks on estimator target that depend on estimator type.
 
-        :param y: 1D numpy array; target for supervised learning
+        :param y: list-like object; target for supervised learning
         :param reset: boolean; whether to prepare the model for a new fit or enable warm start
-        :return: y after applying appropriate checks and transforms
+        :return: numpy array representing y
         """
         raise NotImplementedError("Abstract method Model._validate_y() has no default implementation.")
 
     @staticmethod
     @abstractmethod
-    def _get_compute_classes():  # pragma: no cover
+    def _get_compute_classes(use_tensorflow):  # pragma: no cover
         """Provide classes implementing the set manager and objective function for the model.
 
+        :param use_tensorflow: see docstring of __init__() for details
         :return: subclasses of proset.set_manager.SetManager and proset.objective.Objective
         """
         raise NotImplementedError("Abstract method Model._get_compute_classes() has no default implementation.")
@@ -223,16 +235,20 @@ class Model(BaseEstimator, metaclass=ABCMeta):
             arrays containing the familiarity of each sample
         """
         check_is_fitted(self, attributes="set_manager_")
-        return self._compute_prediction(X=check_array(X), n_iter=n_iter, compute_familiarity=compute_familiarity)
+        return self._compute_prediction(
+            X=check_array(X, **shared.FLOAT_TYPE),
+            n_iter=n_iter,
+            compute_familiarity=compute_familiarity
+        )
 
     @abstractmethod
     def _compute_prediction(self, X, n_iter, compute_familiarity):  # pragma: no cover
         """Compute prediction.
 
-        :param X: see docstring of predict() for details
+        :param X: 2D numpy array of type specified by shared.FLOAT_TYPE; feature matrix
         :param n_iter: see docstring of predict() for details
         :param compute_familiarity: see docstring of predict() for details
-        :return: see docstring of predict() for details
+        :return: as return value of Model.predict()
         """
         raise NotImplementedError("Abstract method Model._get_prediction() has no default implementation.")
 
@@ -241,8 +257,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
 
         :param X: 2D numpy array; feature matrix; sparse matrices or infinite/missing values not supported
         :param y: list-like object; target for supervised learning
-        :param sample_weight: 1D numpy array of positive floats or None; sample weights used for likelihood calculation;
-            pass None to use unit weights
+        :param sample_weight: 1D numpy array of non-negative floats or None; sample weights used for likelihood
+            calculation; pass None to use unit weights
         :param n_iter: non-negative integer, 1D numpy array of non-negative and strictly increasing integers, or None;
             number of batches to use for evaluation; pass None for all batches; pass an array to evaluate for multiple
             values at once
@@ -257,7 +273,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
     def _compute_score(self, X, y, sample_weight, n_iter):  # pragma: no cover
         """Compute score.
 
-        :param X: see docstring of score() for details
+        :param X: 2D numpy array of type specified by shared.FLOAT_TYPE; feature matrix
         :param y: numpy array; target for supervised learning
         :param sample_weight: see docstring of score() for details
         :param n_iter: see docstring of score() for details
@@ -278,8 +294,10 @@ class Model(BaseEstimator, metaclass=ABCMeta):
 
         :param n_iter: non-negative integer, or None; number of batches to use for evaluation; pass None for all
             batches
-        :param train_names: list of strings or None; names for the original training samples in order; these are
-            associated with the prototypes in the report; pass None to use default names 'sample 0', 'sample 1', etc.
+        :param train_names: list of strings, list of lists of strings, or None; names for the original training samples
+            in order; these are associated with the prototypes in the report; if the model was trained using the warm
+            start option, provide a list of lists with sample names per batch up to n_iter or the total number of
+            batches; pass None to use default names 'sample 0', 'sample 1', etc.
         :param include_features: boolean; whether to include information on relevant features
         :param feature_names: list of strings or None; if not None, must have one element per column of features;
             feature names to be used as column headers; pass None to use default names X0, X1, etc.; only used if
@@ -309,6 +327,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         """
         check_is_fitted(self, attributes="set_manager_")
         feature_columns, include_original, scale, offset = self._check_report_input(
+            n_iter=n_iter if n_iter is not None else self.set_manager_.num_batches,
+            train_names=train_names,
             feature_names=feature_names,
             num_features=self.n_features_in_,
             scale=scale,
@@ -333,9 +353,11 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         return report
 
     @staticmethod
-    def _check_report_input(feature_names, num_features, scale, offset, sample_name):
+    def _check_report_input(n_iter, train_names, feature_names, num_features, scale, offset, sample_name):
         """Check input for export() and explain() for consistency and apply defaults.
 
+        :param n_iter: non-negative integer; number of batches to evaluate
+        :param train_names: see docstring of export() for details
         :param feature_names: see docstring of export() for details
         :param num_features: positive integer; number of features
         :param scale: see docstring of export() for details
@@ -344,12 +366,23 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         :return: five return arguments:
             - list of lists of strings; each list contains column names associated with one feature in the report
             - boolean; whether original values need to be included in the report
-            - 1D numpy float array; scale as input or vector of ones if input is None
-            - 1D numpy float array; offset as input or vector of zeros if input is None
+            - 1D numpy array of type specified by shared.FLOAT_TYPE; scale as input or vector of ones if input is None
+            - 1D numpy array of type specified by shared.FLOAT_TYPE; offset as input or vector of zeros if input is None
             - string; sample name as input or default
-            raise an error if a check fails
+            Raises an error if a check fails.
         """
-        feature_names = check_feature_names(
+        if train_names is not None and n_iter > 0:
+            if len(train_names) == 0:
+                raise ValueError("Parameter train_names must not be empty, pass None to use default sample names.")
+            if isinstance(train_names[0], list) and len(train_names) != n_iter:
+                raise ValueError(" ".join([
+                    "Parameter train_names must have as many elements as the number of batches to be evaluated",
+                    "if passing a list."
+                ]))
+            # there is no detailed check whether the list(s) of names have the correct length, as the number of samples
+            # used to create each batch is not available after fitting; if the lists are shorter than the sample indices
+            # for a given batch, report creation fails with an index error later
+        feature_names = shared.check_feature_names(
             num_features=num_features,
             feature_names=feature_names,
             active_features=None
@@ -361,7 +394,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
             "{} similarity".format(feature_name)
         ] for feature_name in feature_names]
         include_original = scale is not None or offset is not None
-        scale, offset = check_scale_offset(num_features=num_features, scale=scale, offset=offset)
+        scale, offset = shared.check_scale_offset(num_features=num_features, scale=scale, offset=offset)
         if sample_name is None:
             sample_name = "new sample"
         return feature_columns, include_original, scale, offset, sample_name
@@ -408,6 +441,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         :return: as return value of _make_prototype_report(); the function determines whether impact needs to be
             computed by checking whether the batch definitions contain the key "similarities"
         """
+        if train_names is not None and isinstance(train_names[0], list):
+            train_names = train_names[batch_index]
         formatted = {
             "batch": batch_index + 1,
             "sample": batch["sample_index"],
@@ -419,7 +454,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         }
         columns = ["batch", "sample", "sample name", "target", "prototype weight"]
         if "similarities" in batch.keys():
-            formatted["similarity"] = np.exp(np.sum(np.log(batch["similarities"] + LOG_OFFSET), axis=1))
+            formatted["similarity"] = np.exp(np.sum(np.log(batch["similarities"] + shared.LOG_OFFSET), axis=1))
             # use sum of logarithms instead of product for numerical stability
             formatted["impact"] = formatted["similarity"] * formatted["prototype weight"]
             columns.extend(["similarity", "impact"])
@@ -441,8 +476,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         :param batches: list as generated by set_manager.SetManager.get_batches()
         :param feature_columns: as first return value of _check_report_input()
         :param include_original: boolean; whether to include original feature values in the report
-        :param scale: see docstring of export() for details; None is not allowed
-        :param offset: see docstring of export() for details; None is not allowed
+        :param scale: as third return value of _check_report_input
+        :param offset: as fourth return value of _check_report_input
         :param active_features: 1D numpy array of non-negative integers; indices of active features across all batches
         :param include_similarities: boolean; whether to include per-feature similarities in the report; if True, the
             information for each non-empty batch needs to contain the key 'similarities'
@@ -475,8 +510,8 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         :param feature_index: positive integer; index of feature
         :param feature_columns: list of strings; as return value of _check_report_input()
         :param include_original: boolean; whether to include original feature values in the report
-        :param scale: see docstring of export() for details; None is not allowed
-        :param offset: see docstring of export() for details; None is not allowed
+        :param scale: as third return value of _check_report_input
+        :param offset: as fourth return value of _check_report_input
         :param include_similarities: boolean; whether to include per-feature similarities in the report
         :return: as one set of columns for the return value of _make_feature_report()
         """
@@ -488,7 +523,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
             if batch is not None:
                 position = np.nonzero(feature_index == batch["active_features"])[0]
                 if len(position) == 0:  # feature is not used by current batch
-                    nan_column = np.NaN * np.zeros(batch["prototype_weights"].shape[0], dtype=float)
+                    nan_column = np.NaN * np.zeros(batch["prototype_weights"].shape[0], **shared.FLOAT_TYPE)
                     new_info = pd.DataFrame({
                         feature_columns[0]: nan_column,
                         feature_columns[1]: nan_column
@@ -545,14 +580,16 @@ class Model(BaseEstimator, metaclass=ABCMeta):
 
         :param X: 2D numpy array; feature matrix having a single row; sparse matrices or infinite/missing values not
             supported
-        :param y: single value; target value for sample to be explained; pass None if true value Is not known
+        :param y: single value; target value for sample to be explained; pass None if true value is not known
         :param n_iter: non-negative integer, or None; number of batches to use for evaluation; pass None for all batches
         :param familiarity: 1D numpy array of non-negative floats or None; if not None, reference values for familiarity
             used to convert absolute familiarity of the new sample to a quantile
         :param sample_name: string or None; name for the new sample; pass None to use name 'new sample'; if both
             sample_name and train_names are None, no names are included in the explanation
-        :param train_names: list of strings or None; names for the original training samples in order; these are
-            associated with the prototypes in the report; pass None to use no names
+        :param train_names: list of strings, list of lists of strings, or None; names for the original training samples
+            in order; these are associated with the prototypes in the report; if the model was trained using the warm
+            start option, provide a list of lists with sample names per batch up to n_iter or the total number of
+            batches; pass None to use default names 'sample 0', 'sample 1', etc.
         :param include_features: boolean; whether to include information on relevant features
         :param feature_names: list of strings or None; if not None, must have one element per column of features;
             feature names to be used as column headers; pass None to use default names X0, X1, etc.; only used if
@@ -589,12 +626,15 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         """
         check_is_fitted(self, attributes="set_manager_")
         feature_columns, include_original, scale, offset, sample_name = self._check_report_input(
+            n_iter=n_iter if n_iter is not None else self.set_manager_.num_batches,
+            train_names=train_names,
             feature_names=feature_names,
             num_features=self.n_features_in_,
             scale=scale,
             offset=offset,
             sample_name=sample_name
         )
+        X = check_array(X, **shared.FLOAT_TYPE)
         batches = self.set_manager_.get_batches(features=X, num_batches=n_iter)
         report = self._make_prototype_report(batches=batches, train_names=train_names, compute_impact=X is not None)
         report = pd.concat([report, self._make_contribution_report(report)], axis=1)
@@ -652,7 +692,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
     ):  # pragma: no cover
         """Format properties of baseline estimator for explain().
 
-        :param X: see docstring of explain() for details
+        :param X: 2D numpy array of type specified by shared.FLOAT_TYPE; feature matrix
         :param y: see docstring of explain() for details
         :param n_iter: see docstring of explain() for details
         :param familiarity: see docstring of explain() for details
@@ -668,7 +708,7 @@ class Model(BaseEstimator, metaclass=ABCMeta):
         raise NotImplementedError("Abstract method Model._get_baseline_for_explain() has no default implementation.")
 
     def shrink(self):
-        """"Reduce internal state representation of a fitted model to active features across all batches.
+        """Reduce internal state representation of a fitted model to active features across all batches.
 
         :return: no return value; adds property active_features_ with feature indices w.r.t. the original training data
         """
@@ -687,10 +727,12 @@ class ClassifierModel(Model):
     def _validate_y(self, y, reset):
         """Perform checks on classification target.
 
-        :param y: 1D numpy array; target for supervised learning
-        :param reset: boolean; whether to prepare the model for a new fit or enable warm start
-        :return: y after applying appropriate checks and transforms; also adds properties label_encoder_ and classes_
-            to the model object
+        :param y: list-like object; target for classification
+        :param reset: see docstring of Model._validate_y() for details
+        :return: 1D numpy integer array representing class labels as integers from 0 to K - 1; the model also gains two
+            new properties:
+            - classes_: 1D numpy array of original class labels
+            - label_encoder_: object of type sklearn.preprocessing.LabelEncoder used to convert class labels to integers
         """
         check_classification_targets(y)
         if reset or not hasattr(self, "label_encoder_"):
@@ -701,20 +743,23 @@ class ClassifierModel(Model):
         return self.label_encoder_.transform(y)
 
     @staticmethod
-    def _get_compute_classes():
+    def _get_compute_classes(use_tensorflow):
         """Provide classes implementing the set manager and objective function for the model.
 
-        :return: subclasses of proset.set_manager.SetManager and proset.objective.Objective
+        :param use_tensorflow: see docstring of Model._get_compute_classes()
+        :return: subclasses of proset.set_manager.SetManager and proset.objective.Objective suitable for classification
         """
-        return ClassifierSetManager, ClassifierObjective
+        if use_tensorflow:
+            return ClassifierSetManager, TfClassifierObjective
+        return ClassifierSetManager, NpClassifierObjective
 
     def _compute_prediction(self, X, n_iter, compute_familiarity):
         """Compute prediction.
 
-        :param X: see docstring of Model.predict() for details
-        :param n_iter: see docstring of Model.predict() for details
-        :param compute_familiarity: see docstring of Model.predict() for details
-        :return: see docstring of Model.predict() for details
+        :param X: see docstring of Model._compute_prediction() for details
+        :param n_iter: see docstring of Model._compute_prediction() for details
+        :param compute_familiarity: see docstring of Model._compute_prediction() for details
+        :return: as return value of Model._compute_prediction()
         """
         prediction = self.set_manager_.evaluate(
             features=X, num_batches=n_iter, compute_familiarity=compute_familiarity
@@ -736,20 +781,24 @@ class ClassifierModel(Model):
     def _compute_score(self, X, y, sample_weight, n_iter):
         """Compute log-likelihood (not multiplied by -1 so it works with sklearn cross-validation).
 
-        :param X: see docstring of Model.score() for details
-        :param y: numpy array; target for supervised learning
-        :param sample_weight: see docstring of Model.score() for details
-        :param n_iter: see docstring of Model.score() for details
-        :return: as return value of Model.score()
+        :param X: see docstring of Model._compute_score() for details
+        :param y: 1D numpy integer array; class labels for classification as integers from 0 to K - 1
+        :param sample_weight: see docstring of Model._compute_score() for details
+        :param n_iter: see docstring of Model._compute_score() for details
+        :return: as return value of Model._compute_score()
         """
-        prediction = self.set_manager_.evaluate(features=X, num_batches=n_iter, compute_familiarity=False)
+        prediction = self.set_manager_.evaluate(
+            features=X,
+            num_batches=n_iter,
+            compute_familiarity=False
+        )
         prediction = [np.squeeze(np.take_along_axis(p, y[:, None], axis=1)) for p in prediction]
         # keep only probability assigned to true class
         if sample_weight is None:
-            prediction = [np.mean(np.log(p + LOG_OFFSET)) for p in prediction]
+            prediction = [np.mean(np.log(p + shared.LOG_OFFSET)) for p in prediction]
         else:
             total_weight = np.sum(sample_weight)
-            prediction = [np.inner(np.log(p + LOG_OFFSET), sample_weight) / total_weight for p in prediction]
+            prediction = [np.sum(np.log(p + shared.LOG_OFFSET) * sample_weight) / total_weight for p in prediction]
         if isinstance(n_iter, np.ndarray):
             return np.array(prediction)
         return prediction[0]
@@ -762,16 +811,17 @@ class ClassifierModel(Model):
             number of batches to use for evaluation; pass None for all batches; pass an array to evaluate for multiple
             values at once
         :param compute_familiarity: boolean; whether to compute the familiarity for each sample
-        :return: 2D numpy array or list of 2D numpy arrays; each row contains the estimated class probabilities for the
-            corresponding row in the feature matrix; if n_iter is integer or None, a single set of predictions is
-            returned as an array; if n_iter is an array, a list of predictions is returned with one element for each
-            element of the array; if compute_familiarity is True, also returns a 1D numpy float array or list of float
-            arrays containing the familiarity of each sample
+        :return: 2D numpy array or list of 2D numpy arrays of type specified by shared.FLOAT_TYPE; each row contains the
+            estimated class probabilities for the corresponding row in the feature matrix; if n_iter is integer or None,
+            a single set of predictions is returned as an array; if n_iter is an array, a list of predictions is
+            returned with one element for each element of the array; if compute_familiarity is True, also returns a 1D
+            numpy float array of type specified by shared.FLOAT_TYPE or list of float arrays containing the familiarity
+            of each sample
         """
         check_is_fitted(self, attributes="set_manager_")
         # noinspection PyUnresolvedReferences
         prediction = self.set_manager_.evaluate(
-            features=check_array(X), num_batches=n_iter, compute_familiarity=compute_familiarity
+            features=check_array(X, **shared.FLOAT_TYPE), num_batches=n_iter, compute_familiarity=compute_familiarity
         )
         if isinstance(n_iter, np.ndarray):
             if compute_familiarity:
@@ -784,7 +834,7 @@ class ClassifierModel(Model):
     def _make_baseline_for_export(self):
         """Format marginal probabilities for export().
 
-        :return: as return value of _make_prototype_report() without columns 'similarity' and 'impact'
+        :return: as return value of Model._make_baseline_for_export()
         """
         return pd.DataFrame({
             "batch": np.NaN,
@@ -807,7 +857,7 @@ class ClassifierModel(Model):
     def _make_contribution_report(self, prototype_report):
         """Format contribution of prototypes to prediction for report.
 
-        :param prototype_report: as return value of _make_prototype_report()
+        :param prototype_report: see docstring of Model._make_contribution_report() for details
         :return: pandas data frame with the following fields:
             - dominant set: 0 or 1; indicates whether the prototype belongs to the dominant set
             - p class <class>: positive float; contribution of prototype to the estimated probability for the given
@@ -838,7 +888,7 @@ class ClassifierModel(Model):
               one row per prototype and one column per class
             - 1D numpy integer array; dominant set indicated by 1, other prototypes by 0
         """
-        contributions = np.zeros((impact.shape[0], marginals.shape[0]), dtype=float)
+        contributions = np.zeros((impact.shape[0], marginals.shape[0]), **shared.FLOAT_TYPE)
         scale = np.sum(impact) + 1.0
         if scale == 1.0:
             # the sample is so far away from the prototypes that the impact of prototypes is below numerical tolerance
@@ -850,7 +900,7 @@ class ClassifierModel(Model):
         rank = np.hstack([0, 1 + np.max(rank) - rank])
         # the marginals have rank 0, all prototypes with the largest impact rank 1, etc.
         extended = np.vstack([marginals / scale, contributions[sort_ix, :]])
-        extended = np.cumsum(np.add.reduceat(extended, indices=find_changes(rank), axis=0), axis=0)
+        extended = np.cumsum(np.add.reduceat(extended, indices=shared.find_changes(rank), axis=0), axis=0)
         # cumulative contribution to probability across ranks
         remainder = 1.0 - np.sum(extended, axis=1)  # remainder of probability missing up to a given rank
         top_two = np.array([np.sort(extended[i, :])[-1:-3:-1] for i in range(extended.shape[0])])
@@ -882,19 +932,19 @@ class ClassifierModel(Model):
     ):
         """Format properties of baseline estimator for explain().
 
-        :param X: see docstring of explain() for details
-        :param y: see docstring of explain() for details
-        :param n_iter: see docstring of explain() for details
-        :param familiarity: see docstring of explain() for details
-        :param sample_name: see docstring of explain() for details
-        :param include_features: see docstring of explain() for details
-        :param active_features: 1D numpy array of non-negative integers; indices of active features across all batches
-        :param feature_columns: as first return value of _check_report_input()
-        :param include_original: boolean; whether to include original feature values in the report
-        :param scale: see docstring of explain() for details; None is not allowed
-        :param offset: see docstring of explain() for details; None is not allowed
-        :return: pandas data frame with the same columns as return value of explain(); the columns describing the
-            contribution of prototypes to the classification result are
+        :param X: see docstring of Model._make_baseline_for_explain() for details
+        :param y: see docstring of Model._make_baseline_for_explain() for details
+        :param n_iter: see docstring of Model._make_baseline_for_explain() for details
+        :param familiarity: see docstring of Model._make_baseline_for_explain() for details
+        :param sample_name: see docstring of Model._make_baseline_for_explain() for details
+        :param include_features: see docstring of Model._make_baseline_for_explain() for details
+        :param active_features: see docstring of Model._make_baseline_for_explain() for details
+        :param feature_columns: see docstring of Model._make_baseline_for_explain() for details
+        :param include_original: see docstring of Model._make_baseline_for_explain() for details
+        :param scale: see docstring of Model._make_baseline_for_explain() for details
+        :param offset: see docstring of Model._make_baseline_for_explain() for details
+        :return: pandas data frame as return value of Model.explain(); the columns describing the contribution of
+            prototypes to the classification result are
             - dominant set: integer; 1 indicates the prototype belongs to the dominant set, 0 that it does not
             - p class <class>: for the new sample, the predicted probability to belong to the given class; for other
               rows, the contribution towards that estimate
@@ -925,7 +975,7 @@ class ClassifierModel(Model):
         ]
         for i in range(self.classes_.shape[0]):
             column_name = "p class {}".format(i)
-            new_column = np.zeros(self.classes_.shape[0] + 1, dtype=float)
+            new_column = np.zeros(self.classes_.shape[0] + 1, **shared.FLOAT_TYPE)
             new_column[0] = probabilities[0, i]
             new_column[i + 1] = self.set_manager_.marginals[i] / (sample_familiarity + 1.0)
             report[column_name] = new_column
